import cterasdk.settings
from ..services import CTERA, client_settings
from ..endpoints import EndpointBuilder
from ...clients.asynchronous import clients

from .. import authenticators

from ...lib.session.core import Session

from ...asynchronous.core import files
from ...asynchronous.core import login
from ...asynchronous.core import cloudfs
from ...asynchronous.core import notifications
from ...asynchronous.core import portals
from ...asynchronous.core import settings
from ...asynchronous.core import users


class Clients:

    def __init__(self, core):
        self.v1 = V1(core)
        self.v2 = V2(core)
        self.io = IO(core)


class V1:

    def __init__(self, core):
        self.ctera = core.default.clone(clients.AsyncExtended, EndpointBuilder.new(core.base, core.context))
        self.api = core.default.clone(clients.AsyncAPI, EndpointBuilder.new(core.base, core.context, '/api'))


class V2:

    def __init__(self, core):
        self.api = core.default.clone(clients.AsyncJSON, EndpointBuilder.new(core.base, core.context, '/v2/api'))


class IO:

    def __init__(self, core):
        self._folders = core.default.clone(clients.AsyncFolders, EndpointBuilder.new(core.base, core.context, '/folders/folders'))
        self._upload = core.default.clone(clients.AsyncUpload, EndpointBuilder.new(core.base, core.context, '/upload/folders'))
        self._webdav = core.default.clone(clients.AsyncWebDAV, EndpointBuilder.new(core.base, core.context, '/webdav'))

    @property
    def upload(self):
        return self._upload.upload

    @property
    def download(self):
        return self._webdav.get

    @property
    def download_zip(self):
        return self._folders.download_zip

    @property
    def builder(self):
        return self._webdav._builder  # pylint: disable=protected-access


class AsyncPortal(CTERA):

    async def __aenter__(self):
        return self

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https, base=None)
        self._default = clients.AsyncClient(EndpointBuilder.new(self.base),
                                            settings=client_settings(cterasdk.settings.sessions.metadata_connector),
                                            authenticator=self._authenticator)
        self._ctera_session = Session(self.host(), self.context)
        self._ctera_clients = Clients(self)

        self.cloudfs = cloudfs.CloudFS(self)
        self.files = files.CloudDrive(self)
        self.notifications = notifications.Notifications(self)
        self.settings = settings.Settings(self)
        self.users = users.Users(self)

    @property
    def default(self):
        return self._default

    @property
    def v1(self):
        return self.clients.v1

    @property
    def v2(self):
        return self.clients.v2

    @property
    def io(self):
        return self.clients.io

    async def login(self, username, password):
        self._before_login()
        await self._login_object.login(username, password)
        await self._ctera_session.async_start_session(self)
        self._after_login()

    async def logout(self):
        if self._ctera_session.connected:
            await self._login_object.logout()
            self._ctera_session.stop_session()
        await self.default.close()

    @property
    def _login_object(self):
        return login.Login(self)

    def _authenticator(self, url):
        return authenticators.core(self.session(), url, self.context)

    async def __aexit__(self, exc_type, exc, tb):
        await self.default.close()


class AsyncGlobalAdmin(AsyncPortal):

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https)
        self.portals = portals.Portals(self)

    @property
    def context(self):
        return 'admin'

    async def impersonate(self, username, tenant):
        """
        Impersonate a Portal user

        :param str username: Username
        :param str tenant: Tenant
        """
        ctera_ticket = await self.users.generate_ticket(username, tenant)
        user = AsyncServicesPortal(f'{tenant}.{await self.settings.global_settings.dns_suffix}', self.port())
        await user.sso(ctera_ticket)
        return user


class AsyncServicesPortal(AsyncPortal):

    @property
    def context(self):
        return 'ServicesPortal'

    async def sso(self, ctera_ticket):
        """
        Login using a Portal ticket

        :param str ctera_ticket: SSO Ticket.
        """
        await self._login_object.sso(ctera_ticket)
        await self._ctera_session.async_start_session(self)
        await self.v1.api.web_session()



from .common import Object
from .convert import tojsonstr


class CTERAException(Exception):

    def __init__(self, message=None, instance=None, **kwargs):
        super().__init__()
        self.classname = self.__class__.__name__
        self.join(instance)
        if message is not None:
            self.message = message
        self.put(**kwargs)

    def put(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def join(self, instance=None):
        if instance is not None:
            self.__dict__ = instance.__dict__.copy()

    def __str__(self):
        return tojsonstr(self)


class ContextError(CTERAException):
    """API Invocation Context Error"""


class SessionExpired(CTERAException):
    """Raised on Session Expiration"""

    def __init__(self):
        super().__init__('Session expired.')


class NotLoggedIn(CTERAException):
    """Raised on No Session"""

    def __init__(self):
        super().__init__('Not logged in.')


class ClientResponseException(CTERAException):

    def __init__(self, error_object):
        super().__init__('An error occurred while processing the HTTP request.', error_object)


class NotificationsError(CTERAException):

    def __init__(self, cloudfolders, cursor):
        super().__init__('An error occurred while trying to retrieve notifications.', cloudfolders=cloudfolders, cursor=cursor)


class ObjectNotFoundException(CTERAException):

    def __init__(self, message, object_ref, **kwargs):
        super().__init__(message, None, **kwargs)
        self.response = Object()
        self.response.code = 404
        self.response.reason = 'Not Found'
        self.response.body = Object()
        self.response.body.rc = 1
        self.response.body.msg = f"Object '{object_ref}' not found"


class PythonVersionException(CTERAException):

    def __init__(self, version):
        super().__init__('You are running an unsupported version of Python', None, version=version)


class InputError(CTERAException):

    def __init__(self, message, expression, options):
        CTERAException.__init__(self, message, None, expression=expression, options=options)


class ConsentException(CTERAException):
    """Consent Exception"""



import logging
from ...cio.common import encode_request_parameter
from ...cio import edge as fs
from ...cio import exceptions


logger = logging.getLogger('cterasdk.edge')


def listdir(edge, path):
    with fs.listdir(path) as param:
        return edge.api.execute('/status/fileManager', 'listPhysicalFolders', param)


def mkdir(edge, path):
    with fs.makedir(path) as param:
        edge.io.mkdir(param)
    return path.absolute


def makedirs(edge, path):
    directories = path.parts
    for i in range(1, len(directories) + 1):
        path = fs.EdgePath(path.scope, '/'.join(directories[:i]))
        try:
            mkdir(edge, path)
        except exceptions.RestrictedPathError:
            logger.warning('Creating a folder in the specified location is forbidden: %s', path.reference.as_posix())


def copy(edge, path, destination=None, overwrite=False):
    with fs.copy(path, destination) as (src, dst):
        edge.io.copy(src, dst, overwrite=overwrite)


def move(edge, path, destination=None, overwrite=False):
    with fs.move(path, destination) as (src, dst):
        edge.io.move(src, dst, overwrite=overwrite)


def remove(edge, *paths):
    for path in fs.delete_generator(*paths):
        edge.io.delete(path.absolute)


def handle(path):
    """
    Create function to retrieve file handle.

    :param cterasdk.cio.edge.EdgePath path: Path to file.
    :returns: Callable function to retrieve file handle.
    :rtype: callable
    """
    def wrapper(edge):
        """
        Get file handle.

        :param cterasdk.objects.synchronous.edge.Edge edge: Edge Filer object.
        """
        with fs.handle(path) as param:
            return edge.io.download(param)
    return wrapper


def handle_many(directory, *objects):
    """
    Create function to retrieve zip archive

    :param cterasdk.cio.edge.EdgePath directory: Path to directory.
    :param args objects: List of files and folders.
    :returns: Callable function to retrieve file handle.
    :rtype: callable
    """
    def wrapper(edge):
        """
        Upload file from metadata and file handle.

        :param cterasdk.objects.synchronous.edge.Edge edge: Edge Filer object.
        :param str name: File name.
        :param object handle: File handle.
        """
        with fs.handle_many(directory, objects) as param:
            return edge.io.download_zip('/admingui/api/status/fileManager/zip', encode_request_parameter(param))
    return wrapper


def upload(name, destination, fd):
    """
    Create upload function

    :param str name: File name.
    :param cterasdk.cio.edge.EdgePath destination: Path to directory.
    :param object fd: File handle.
    :returns: Callable function to start the upload.
    :rtype: callable
    """
    def wrapper(edge):
        """
        Upload file from metadata and file handle.

        :param cterasdk.objects.synchronous.edge.Edge edge: Edge Filer object.
        """
        with fs.upload(name, destination, fd) as param:
            return edge.io.upload('/actions/upload', param)
    return wrapper



from ..base_command import BaseCommand
from ...cio.edge import EdgePath
from ...lib import FileSystem
from . import io


class FileBrowser(BaseCommand):
    """ Edge Filer File Browser APIs """

    def __init__(self, edge):
        super().__init__(edge)
        self._filesystem = FileSystem.instance()

    def listdir(self, path):
        """
        List Directory

        :param str path: Path
        """
        return io.listdir(self._edge, path)

    def handle(self, path):
        """
        Get File Handle.

        :param str path: Path to a file
        """
        handle_function = io.handle(self.normalize(path))
        return handle_function(self._edge)

    def handle_many(self, directory, *objects):
        """
        Get a Zip Archive File Handle.

        :param str directory: Path to a folder
        :param args objects: List of files and folders
        """
        handle_many_function = io.handle_many(self.normalize(directory), *objects)
        return handle_many_function(self._edge)

    def download(self, path, destination=None):
        """
        Download a file

        :param str path: The file path on the Edge Filer
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        directory, name = self.determine_directory_and_filename(path, destination=destination)
        handle = self.handle(path)
        return self._filesystem.save(directory, name, handle)

    def download_as_zip(self, target, objects, destination=None):
        """
        Download a list of files and/or directories from a cloud folder as a ZIP file

        .. warning:: The list of files is not validated. The ZIP file will include only the existing  files and directories

        :param str target: Path to a directory
        :param list[str] objects: List of files and/or directories in the cloud folder to download
        :param str,optional destination:
         File destination, if it is a directory, the filename will be calculated, defaults to the default directory
        """
        directory, name = self.determine_directory_and_filename(target, objects, destination=destination, archive=True)
        handle = self.handle_many(target, *objects)
        return self._filesystem.save(directory, name, handle)

    def upload(self, name, destination, handle):
        """
        Upload from file handle.

        :param str name: File name.
        :param str destination: Path to remote directory.
        :param object handle: Handle.
        """
        upload_function = io.upload(name, self.normalize(destination), handle)
        return upload_function(self._edge)

    def upload_file(self, path, destination):
        """
        Upload a file.

        :param str path: Local path
        :param str destination: Remote path
        """
        metadata = self._filesystem.properties(path)
        with open(path, 'rb') as handle:
            response = self.upload(metadata['name'], destination, handle)
        return response

    def mkdir(self, path):
        """
        Create a new directory

        :param str path: Directory path
        """
        return io.mkdir(self._edge, self.normalize(path))

    def makedirs(self, path):
        """
        Create a directory recursively

        :param str path: Directory path
        """
        return io.makedirs(self._edge, self.normalize(path))

    def copy(self, path, destination=None, overwrite=False):
        """
        Copy a file or a folder

        :param str path: Source file or folder path
        :param str destination: Destination folder path
        :param bool,optional overwrite: Overwrite on conflict, defaults to False
        """
        if destination is None:
            raise ValueError('Copy destination was not specified.')
        return io.copy(self._edge, self.normalize(path), self.normalize(destination), overwrite)

    def move(self, path, destination=None, overwrite=False):
        """
        Move a file or a folder

        :param str path: Source file or folder path
        :param str destination: Destination folder path
        :param bool,optional overwrite: Overwrite on conflict, defaults to False
        """
        if destination is None:
            raise ValueError('Move destination was not specified.')
        return io.move(self._edge, self.normalize(path), self.normalize(destination), overwrite)

    def delete(self, path):
        """
        Delete a file

        :param str path: File path
        """
        return io.remove(self._edge, self.normalize(path))

    def determine_directory_and_filename(self, p, objects=None, destination=None, archive=False):
        """
        Determine location to save file.

        :param str p: Path.
        :param list[str],optional objects: List of files or folders
        :param str,optional destination: Destination
        :param bool,optional archive: Compressed archive
        :returns: Directory and file name
        :rtype: tuple[str]
        """
        directory, name = None, None
        if destination:
            directory, name = self._filesystem.split_file_directory(destination)
        else:
            directory = self._filesystem.downloads_directory()

        if not name:
            normalized = self.normalize(p)
            if archive:
                name = self._filesystem.compute_zip_file_name(normalized.absolute, objects)
            else:
                name = normalized.name
        return directory, name

    @staticmethod
    def normalize(path):
        return EdgePath('/', path)



import logging
from . import async_requests
from ..common import utils


logger = logging.getLogger('cterasdk.http')


class CookieJar:

    def __init__(self, cookies):
        self._cookies = cookies

    def get(self, key):
        return self._cookies.get(key)

    def update(self, cookies, response_url):
        self._cookies.update_cookies(cookies, response_url)

    def clear(self):
        self._cookies.clear()


class PersistentHeaders:
    """Headers to include in every request"""

    def __init__(self):
        self._headers = {}

    @property
    def all(self):
        return self._headers

    def persist_response_header(self, response, header):
        """
        Persist header from response object.

        :param cterasdk.clients.base.BaseResponse response: Response object
        :param str header: Header name
        """
        value = response.headers.get(header, None)
        if value:
            self.persist_headers({header: value})
        else:
            logger.debug("Could not find header: '%s' in response", header)

    def persist_headers(self, headers):
        self._headers.update(headers)


class BaseClient:
    """Base Client"""

    def __init__(self, builder=None, session=None, settings=None, authenticator=None):
        """
        Initialize a Client

        :param builder: Endpoint builder.
        :param ,optional session: Session.
        :param ,optional settings: Client Session Settings.
        :param ,optional authenticator: Authenticator function.
        """
        self._headers = PersistentHeaders()
        self._authenticator = authenticator
        self._builder = builder
        self._session = session if session else async_requests.Session(settings)

    def clone(self, definition, builder=None, authenticator=None):
        """
        Clone a Client

        :param class definition: Class definition.
        :param ,optional builder: Endpoint builder.
        :param ,optional authenticator: Authenticator function.
        """
        return definition(
            builder if builder is not None else self._builder,
            self._session,
            None,
            authenticator if authenticator is not None else self._authenticator,
        )

    @property
    def cookies(self):
        return CookieJar(self._session.cookies)

    @property
    def headers(self):
        return self._headers

    def join_headers(self, request):
        request.kwargs['headers'] = utils.merge(request.kwargs.get('headers', None), self.headers.all)
        return request

    @property
    def baseurl(self):
        return self._builder()

    def request(self, request, *, on_response=None):
        return self._request(request, on_response=on_response)

    async def async_request(self, request, *, on_response=None):
        return await self._request(request, on_response=on_response)

    async def close(self):
        await self._session.close()

    def __str__(self):
        return f"({self.__class__.__name__} client at {hex(hash(self))}, baseurl={self.baseurl})"


class BaseResponse:
    """Base Response Object"""

    def __init__(self, response):
        self._response = response

    @property
    def method(self):
        return self._response.method

    @property
    def ok(self):
        return self._response.ok

    @property
    def reason(self):
        return self._response.reason

    @property
    def url(self):
        return self._response.url

    @property
    def real_url(self):
        return self._response.real_url

    @property
    def status(self):
        return self._response.status

    @property
    def headers(self):
        return self._response.headers

    def raise_for_status(self):
        return self._response.raise_for_status()



import logging
from contextlib import contextmanager
from ..common import Object
from . import common, exceptions


logger = logging.getLogger('cterasdk.edge')


class EdgePath(common.BasePath):
    """Path for CTERA Edge Filer"""

    @staticmethod
    def instance(scope, reference):
        return EdgePath(scope, reference)


@contextmanager
def listdir(path):
    param = Object()
    param.path = path
    logger.info('Listing directory: %s', path)
    yield param


@contextmanager
def makedir(path):
    directory = path.absolute
    logger.info('Creating directory: %s', path.absolute)
    try:
        yield path.absolute
    except exceptions.CTERAException as error:
        try:
            accept_response(error.response.message.msg, directory)
        except exceptions.ResourceExistsError:
            logger.info('Directory already exists: %s', directory)
    logger.info('Directory created: %s', directory)


@contextmanager
def copy(path, destination):
    destination = destination.join(path.name).absolute
    logger.info('Copying: %s to: %s', path.absolute, destination)
    yield path.absolute, destination


@contextmanager
def move(path, destination):
    destination = destination.join(path.name).absolute
    logger.info('Moving: %s to: %s', path.absolute, destination)
    yield path.absolute, destination


def delete_generator(*paths):
    for path in paths:
        logger.info('Deleting item: %s', path.absolute)
        yield path


@contextmanager
def handle(path):
    logger.info('Getting file handle: %s', path.reference)
    yield path.absolute


@contextmanager
def handle_many(directory, objects):
    param = Object()
    param.paths = ['/'.join([directory.absolute, item]) for item in objects]
    param.snapshot = Object()
    param._classname = 'BackupRepository'  # pylint: disable=protected-access
    param.snapshot.location = 1
    param.snapshot.timestamp = None
    param.snapshot.path = None
    logger.info('Getting directory handle: %s', directory.reference)
    yield param


@contextmanager
def upload(name, destination, fd):
    param = dict(
        name=name,
        fullpath=f'{destination.absolute}/{name}',
        filedata=fd
    )
    logger.info('Uploading: %s to: %s', name, destination.reference)
    yield param


def accept_response(response, reference):
    error = {
        "File exists": exceptions.ResourceExistsError(),
        "Creating a folder in this location is forbidden": exceptions.RestrictedPathError(),
    }.get(response, None)
    try:
        if error:
            raise error
    except exceptions.ResourceExistsError as error:
        logger.warning('Resource already exists: a file or folder with this name already exists. %s', {'path': reference})
        raise error
    except exceptions.RestrictedPathError as error:
        logger.error('Creating a folder in the specified location is forbidden. %s', {'name': reference})
        raise error
